import sys,os
import pprint
import ResourceManager
from NeverFile import NeverFile

class ERFKey:
    def __init__(self,n='',i=0,t=0,o=0,s=0):
        self.name = n
        self.id = i
        self.type = t
        self.offset = o
        self.size = s
        self.contents = None

    def getPrintableName(self):
        return self.name.strip('\0')
    
    def __str__(self):
        stringtype = ResourceManager.extensionFromResType(self.type)
        s = 'name: ' + self.name + ' id: ' + `self.id` + ' type: ' + stringtype\
            + ' offset: ' + `self.offset` + ' size: ' + `self.size`
        return s

    def __repr__(self):
        return 'ERFKey(' + `self.name` + ',' + `self.id` + ',' + `self.type`\
               + ',' + `self.offset` + ',' + `self.size` + ')'

class ERFFile(NeverFile):
    """A class encapsulating a Bioware ERF file.
    This class contains two sets of data structures. One set is a flat
    structure that closely corresponds to the actual ERF file data as it
    is stored on disk. The other is a tree that corresponds to the conceptual
    structure of an ERF file. Given a file, this class first reads most
    of the flat structures from the file, and then constructs the tree in
    memory (while still reading some of the flat structure). It will delay
    reading the actual contents of the ERF file files until they are requested.
    Once build, clients can work with and modify the tree structure. When asked
    to save back to a file, this class will flatten the tree structures into
    its flat structures, then saves these to the file."""

    def __init__(self):
        NeverFile.__init__(self)
        self.filename = ""
        self.readFileHandle = None
        self.writeFileHandle = None
    
        self.languageCount = 0
        self.localizedStringSize = 0
        self.entryCount = 0
        self.offsetToLocalizedString = 0
        self.offsetToKeyList = 0
        self.offsetToResourceList = 0
        self.buildYear = 0
        self.buildDay = 0
        self.descriptionRef = ""

        self.localizedStrings = {}
        
        self.entriesByExtension = {}
        self.entriesByNameAndExtension = {}
    
    def headerFromFile(self,f):
        NeverFile.headerFromFile(self,f)
        self.languageCount = self.dataHandler.readIntFile(f)
        self.localizedStringSize = self.dataHandler.readIntFile(f)
        self.entryCount = self.dataHandler.readIntFile(f)
        self.offsetToLocalizedString = self.dataHandler.readIntFile(f)
        self.offsetToKeyList = self.dataHandler.readIntFile(f)
        self.offsetToResourceList = self.dataHandler.readIntFile(f)
        self.buildYear = self.dataHandler.readIntFile(f)
        self.buildDay = self.dataHandler.readIntFile(f)
        self.descriptionRef = f.read(4)
        f.seek(116,1) #skip header padding

    def recalculateParams(self):
        """recalculate the offsets and sizes to be written to a file"""
        self.localizedStringSize = self.calcLocalizedStringsSize()
        self.languageCount = len(self.localizedStrings.keys())
        self.entryCount = len(self.entriesByNameAndExtension)
        offset = self.offsetToLocalizedString # constant header size
        offset += self.localizedStringSize
        self.offsetToKeyList = offset
        offset += self.calcKeyListSize()
        self.offsetToResourceList = offset

    def headerToFile(self):
        """output the header info of the ERF file"""
        self.writeFileHandle.write(self.type)
        self.writeFileHandle.write(self.version)
        self.dataHandler.writeIntFile(self.languageCount,self.writeFileHandle)
        self.dataHandler.writeIntFile(self.localizedStringSize,self.writeFileHandle)
        self.dataHandler.writeIntFile(self.entryCount,self.writeFileHandle)
        self.dataHandler.writeIntFile(self.offsetToLocalizedString,self.writeFileHandle)
        self.dataHandler.writeIntFile(self.offsetToKeyList,self.writeFileHandle)
        self.dataHandler.writeIntFile(self.offsetToResourceList,self.writeFileHandle)
        self.dataHandler.writeIntFile(self.buildYear,self.writeFileHandle)
        self.dataHandler.writeIntFile(self.buildDay,self.writeFileHandle)
        self.writeFileHandle.write(self.descriptionRef)
        self.writeFileHandle.write(116*'\0')
            
    def readLocalizedStrings(self):
        """read the ERF localized string file section"""
        self.readFileHandle.seek(self.offsetToLocalizedString)
        for i in range(self.languageCount):
            (langID,s) = self.dataHandler.readLocalizedString(self.readFileHandle)
            self.localizedStrings[langID] = s

    def writeLocalizedStrings(self):
        """write the ERF localized string file section"""
        self.writeFileHandle.seek(self.offsetToLocalizedString)
        for s in self.localizedStrings:
            self.dataHandler.writeLocalizedStringFile(s,self.localizedStrings[s],self.writeFileHandle)
        
    def calcLocalizedStringsSize(self):
        """calculate the size the ERF localized string section will take"""
        s = 0
        for lID in self.localizedStrings:
            s += 8 #langID and size
            s += len(self.localizedStrings[lID])
        return s
    
    def readKeysAndResourceList(self):
        """read the key and resources ERF sections"""
        self.readFileHandle.seek(self.offsetToKeyList)
        keys = []
        for i in range(self.entryCount):
            k = ERFKey()
            k.name = self.dataHandler.readResRef(self.readFileHandle)#.strip('\0')
            k.id = self.dataHandler.readIntFile(self.readFileHandle)
            k.type = self.dataHandler.readWordFile(self.readFileHandle)
            self.readFileHandle.seek(2,1) #unused
            keys.append(k)
        self.readFileHandle.seek(self.offsetToResourceList)
        for k in keys:
            k.offset = self.dataHandler.readIntFile(self.readFileHandle)
            k.size = self.dataHandler.readIntFile(self.readFileHandle)
            #print 'found entry of type',self.typeToText(k.type)
            if k.type in self.entriesByExtension.keys():
                self.entriesByExtension[k.type].append(k)
            else:
                self.entriesByExtension[k.type] = [k]
            if (k.name,k.type) in self.entriesByExtension.keys():
                print 'error, ',(k.name,k.type),'already present as key'
            self.entriesByNameAndExtension[(k.name,k.type)] = k

    def writeKeys(self):
        """write the key section of an ERF file"""
        self.writeFileHandle.seek(self.offsetToKeyList)
        entryList = [(int(x.id),x) for x in self.entriesByNameAndExtension.values()]
        entryList.sort()
        for e in entryList:
            k = e[1]
            self.dataHandler.writeResRef(k.name,self.writeFileHandle)
            self.dataHandler.writeIntFile(k.id,self.writeFileHandle)
            self.dataHandler.writeWordFile(k.type,self.writeFileHandle)
            self.writeFileHandle.write(2*'\0')

    def writeResourceList(self):
        """write the resource list section of an ERF file.
        Note that this should only be done AFTER the resources have been written,
        because we do not know their sizes and locations beforehand."""        
        self.writeFileHandle.seek(self.offsetToResourceList)
        entryList = [(int(x.id),x) for x in self.entriesByNameAndExtension.values()]
        entryList.sort()
        for e in entryList:
            k = e[1]
            self.dataHandler.writeIntFile(k.offset,self.writeFileHandle)
            self.dataHandler.writeIntFile(k.size,self.writeFileHandle)
    
    def calcKeyListSize(self):
        """calculate the size the key list will take up"""
        #the toolset writes lots of extra 0 bytes after the keys for some
        #reason, 8 for each entry. The game will happily load modules that
        #don't have this padding, but the toolset complains, so I'm now
        #padding, too, to make it shut up.
        return len(self.entriesByNameAndExtension) * 32 #24 is enough

    def calcResourceListSize(self):
        """calculate the size the resource list will take up"""
        return len(self.entriesByNameAndExtension) * 8
        
    def fromFile(self,fname):
        """initialize an ERF file from a file (does not initialize the files contained
        in this ERF file)."""
        self.filename = fname
        self.readFileHandle = NeverFile.fromFile(self,fname)
        self.readLocalizedStrings()
        self.readKeysAndResourceList()
            
    def writeEntries(self):
        """write the actual file entries of this ERF file. This method
        will ask the contents to write themselves by calling 'toFile' on
        them, because they may have changed without our knowledge. However,
        if the contents have never been read, this method will read them
        from the read file and write them out to the file being written,
        meaning that we can never write directly to the file we're reading
        (which is a good thing)."""
        offset = self.offsetToResourceList + self.calcResourceListSize()
        self.writeFileHandle.seek(offset)
        entryList = [(int(x.id),x) for x in self.entriesByNameAndExtension.values()]
        entryList.sort()
        for e in entryList:
            entry = e[1]
            if entry.contents:
                contents = self.getEntryContents(entry)
                contents.toFile(self.writeFileHandle,offset)
            else:
                contents = self.getRawEntryContents(entry)
                self.writeFileHandle.write(contents)
            entry.offset = offset
            entry.size = self.writeFileHandle.tell() - entry.offset
            offset += entry.size

    def saveToReadFile(self):
        """this method simulates writing back out to the file that this ERF
        file was read from. It actually writes a temp file, then moves this
        file to overwrite the old read file."""
        #the next line generates a warning, but I don't know how to
        #do the following
        #rename with tmpfile
        tmp = os.tmpnam()
        self.toFile(tmp)
        self.close()
        print 'renaming',self.writeFileHandle.name,self.readFileHandle.name
        if os.name == 'nt' and os.path.exists(self.readFileHandle.name):
            os.remove(self.readFileHandle.name)
        os.rename(self.writeFileHandle.name,self.readFileHandle.name)
        self.reOpenRead()
        
    def toFile(self,fname=''):
        """write this ERF to a file"""
        if len(fname) > 0:
            if fname == self.readFileHandle.name:
                print 'error, ERFFile cannot write to file it is reading from'
                return
            self.writeFileHandle = open(fname,'wb')
        self.recalculateParams()
        self.headerToFile()
        self.writeLocalizedStrings()
        self.writeKeys()
        self.writeEntries() #do this first, so sizes & offsets are correct
        self.writeResourceList()
        self.writeFileHandle.flush()
        self.writeFileHandle.close()
        self.readFileHandle = open(self.filename,'rb')

    def reassignReadFile(self,fname):
        self.filename = fname
        self.readFileHandle = open(self.filename,'rb')

    def addFile(self,fname):
        """add another ERF file's entries to this one -
        skips localized strings and does not replace entries
        already present (as determined by their key).
        Note that this will save the file."""
        other = ERFFile()
        other.fromFile(fname)

        #a class to give raw contents a 'toFile' method for writing
        class RawContentWrapper:
            def __init__(self,erf,entry):
                self.entry = entry
                self.erf = erf

            def toFile(self,f,offset):
                f.seek(offset)
                f.write(self.erf.getRawEntryContents(self.entry))
        
        #don't use the localized strings from the file being added

        #update the entry list by adding entries from the other
        #file if they don't yet exist
        for (name,type),entry in other.entriesByNameAndExtension.iteritems():
            if (name,type) not in self.entriesByNameAndExtension.keys():
                entry.contents = RawContentWrapper(other,entry)
                entry.id = len(self.entriesByNameAndExtension)
                self.entriesByNameAndExtension[(name,type)] = entry
                if type in self.entriesByExtension.keys():
                    self.entriesByExtension[type].append(entry)
                else:
                    self.entriesByExtension[type] = [entry]

        #transfer contents to original file
        self.saveToReadFile()
        other.close()
        #force all entry contents to be re-interpreted
        for entry in self.entriesByNameAndExtension.values():
            entry.contents = None
    
        

    def close(self):
        """close all the read and write file this class might have opened."""
        if self.readFileHandle:
            self.readFileHandle.close()
        if self.writeFileHandle:
            self.writeFileHandle.close()

    def reOpenRead(self):
        """re-open the read file of this class in case it was closed."""
        if self.readFileHandle:
            self.readFileHandle = open(self.readFileHandle.name,'rb')
    
    def getReadFileName(self):
        """get the filename this class was read from."""
        return self.readFileHandle.name
    
    def getEntriesWithExtension(self,ext):
        """get all the entries that have a given extension.
        @param ext the extension to look up
        @return a list of ERFKey objects"""
        type = ResourceManager.resTypeFromExtension(ext)
        return self.entriesByExtension.get(type, [])

    def getEntryByNameAndExtension(self,name,ext):
        """get an entry with a specific name and extension (that's
        equivalent to looking up by key).
        @param name the name to look up
        @param ext the extension to look up
        @return the ERFKey with this name and extension, None, if not in this file"""
        name += (16-len(name))*'\0'
        type = ResourceManager.resTypeFromExtension(ext)
        return self.entriesByNameAndExtension.get((name,type), None)

    def getRawEntryContents(self,entry):
        """get the raw contents of a given entry as a byte string"""
        self.readFileHandle.seek(entry.offset)
        return self.readFileHandle.read(entry.size)
    
    def getEntryContents(self,entry):
        """get the contents of an entry as interpreted by the ResourceManager.
        This will not re-read them if they've already been read."""
        if not entry.contents:
            raw = self.getRawEntryContents(entry)
            c = ResourceManager.interpretResourceContents((entry.name,
                                                           entry.type),
                                                          raw)
            entry.contents = c
        return entry.contents

    def setEntryContents(self,entry,contents):
        """set the contents of an entry. Does a simple set."""
        entry.contents = contents
        
    def getKeyList(self):
        """Get a list of all keys stored in this ERF file."""
        return self.entriesByNameAndExtension.keys()

    def getResourceData(self,key):
        """get the raw data of an entry corresponding to the given key.
        Just like getRawEntryContents, but with a key interface."""
        extension = ResourceManager.extensionFromResType(key[1])
        return self.getRawEntryContents(self.getEntryByNameAndExtension(key[0],
                                                                        extension))

    def getResource(self,key):
        """Get the contents of an entry specified by a key (Just like
        getEntryContents, but with a key)."""
        extension = ResourceManager.extensionFromResType(key[1])
        return self.getEntryContents(self.getEntryByNameAndExtension(key[0],
                                                                     extension))

    def getRawResource(self,key):
        """Get the raw bytes of an entry specified by a key"""
        extension = ResourceManager.extensionFromResType(key[1])
        return self.getRawEntryContents(self.getEntryByNameAndExtension(key[0],
                                                                     extension))
    
    def __str__(self):
        s =  'ERFFile ' + self.filename + '\n'
        s += self.type + ' ' + self.version + '\n'
        s += `self.languageCount` + ' ' + `self.localizedStringSize` + ' '\
             + `self.entryCount` + '\n'
        s += `self.offsetToLocalizedString` + ' ' + `self.offsetToKeyList`\
             + ' ' + `self.offsetToResourceList` + '\n'
        year = self.buildYear + 1900
        s += `year` + ' ' + `self.buildDay` + '\n'
        s += self.descriptionRef + '\n'
        s += pprint.PrettyPrinter().pformat(f.localizedStrings)
        s += pprint.PrettyPrinter().pformat(f.entriesByNameAndExtension)
        return s
        
if __name__ == "__main__":
    if(len(sys.argv) >= 2):
        f = ERFFile()
        f.fromFile(sys.argv[1])
        print f
        if len(sys.argv) >= 3:
            f.toFile(sys.argv[2])
        #areas = f.getAreas()
        #players = f.getEntriesOfType('GIC')
        #profile.run('fac = f.getEntryContents(players[0])','prof')
        #print fac
        #f.toFile(sys.argv[1] + '.copy')
        #f.fromFile(sys.argv[1] + '.copy')
        #print f
        #pprint.PrettyPrinter().pprint(f.localizedStrings)
        #pprint.PrettyPrinter().pprint(f.entriesByNameAndExtension)
    else:
        print 'usage:',sys.argv[0],'<filename>'



            
            
        
